{*******************************************************}
{                                                       }
{         Delphi VCL Components for                     }
{                 selecting/creating a folder using     }
{                 Windows95/NT style                    }
{                 version 1.4                           }
{                                                       }
{                 Supports                              }
{                 - file notification for fixed and     }
{                   RAM-drives (optional)               }
{                 - network browser                     }
{                                                       }
{         Copyright (c) 1998-2000 Konstantin Polyakov   }
{                 mailto: kpolyakov@mail.ru             }
{                 FIDO: 2:5030/542.251                  }
{                                                       }
{*******************************************************}
unit DirTree;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, CommCtrl,
  ShellAPI,
{$IFNDEF VER90}
  ShlObj,
{$ENDIF}
  StdCtrls, FileCtrl;

{$IFDEF VER90}
//---------  Instead of ShlObj.DCU (for D2 Compatibility) --------
type

  PSHItemID = ^TSHItemID;
  TSHItemID = packed record           { mkid }
    cb: Word;                         { Size of the ID (including cb itself) }
    abID: array[0..0] of Byte;        { The item ID (variable length) }
  end;

  PItemIDList = ^TItemIDList;
  TItemIDList = packed record         { idl }
     mkid: TSHItemID;
  end;

const
{ registry entries for special paths are kept in : }

  CSIDL_DESKTOP                       = $0000;
  CSIDL_PROGRAMS                      = $0002;
  CSIDL_CONTROLS                      = $0003;
  CSIDL_PRINTERS                      = $0004;
  CSIDL_PERSONAL                      = $0005;
  CSIDL_FAVORITES                     = $0006;
  CSIDL_STARTUP                       = $0007;
  CSIDL_RECENT                        = $0008;
  CSIDL_SENDTO                        = $0009;
  CSIDL_BITBUCKET                     = $000a;
  CSIDL_STARTMENU                     = $000b;
  CSIDL_DESKTOPDIRECTORY              = $0010;
  CSIDL_DRIVES                        = $0011;
  CSIDL_NETWORK                       = $0012;
  CSIDL_NETHOOD                       = $0013;
  CSIDL_FONTS                         = $0014;
  CSIDL_TEMPLATES                     = $0015;
  CSIDL_COMMON_STARTMENU              = $0016;
  CSIDL_COMMON_PROGRAMS               = $0017;
  CSIDL_COMMON_STARTUP                = $0018;
  CSIDL_COMMON_DESKTOPDIRECTORY       = $0019;
  CSIDL_APPDATA                       = $001a;
  CSIDL_PRINTHOOD                     = $001b;

function SHGetSpecialFolderLocation(hwndOwner: HWND; nFolder: Integer;
  var ppidl: PItemIDList): HResult; stdcall;
{$ENDIF}
//------------------------------------------------------------------

const
  SErrAllocHandle = '   ';

type

  TFileAttr = (ftReadOnly, ftHidden, ftSystem, ftVolumeID, ftDirectory,
    ftArchive, ftNormal);
  TFileType = set of TFileAttr;

  TTVNewEditCancelEvent = procedure( Sender: TObject;
    Node: TTreeNode; var Delete: Boolean) of object;

  TTVBeginCreatingEvent = procedure( Sender: TObject;
                                                                 Node: TTreeNode) of object;
  TTVEndCreatingEvent = procedure( Sender: TObject;
                                                                 Node: TTreeNode) of object;

{------------------------------------------------------------
  Fnugry File Notify Component
  Copyright (C) 1997 Gleb Yourchenko
  Version 1.0.1.1

  Contact:
  Gleb Yourchenko
  E-mail: eip__@hotmail.com
------------------------------------------------------------}
  EFileNotificationError = class(Exception);

  TNotificationOption = (fnoFile, fnoFolder, fnoAttr,
    fnoSize, fnoTime, fnoWatchSubtree);

  TNotificationOptions = set of TNotificationOption;

  TFileNotifyEvent = procedure(Sender :TObject;
    Action :TNotificationOption) of object;

  TFileNotify = class;

{**********************************************************
                FILE NOTIFICATION THREAD
**********************************************************}
  TNotificationThread = class(TThread)
  private
    FOwner             :TFileNotify;
    FHandles           :array[0..7] of THandle;
    FHandleFilters     :array[0..7] of TNotificationOption;
    FHandleCount       :Integer;
    FActiveFilter      :TNotificationOption;
    procedure ReleaseHandle;
    procedure AllocateHandle;
  protected
    procedure Execute; override;
    procedure Notify;
  public
    constructor Create(Owner :TFileNotify);
    destructor Destroy; override;
    procedure Reset;
  end;

{**********************************************************
                FILE NOTIFY
**********************************************************}
  TFileNotify = class
  private
    FThread      :TNotificationThread;
    FAsyncNotify :Boolean;
    FFolder      :String;
    FOptions     :TNotificationOptions;
    FOnNotify    :TFileNotifyEvent;
    procedure SetOptions(const Value :TNotificationOptions);
    procedure SetRootFolder(const Value :String);
    procedure SetAsyncNotify(Value :Boolean);
  protected
    procedure Notify(Action :TNotificationOption); virtual;
        procedure Prepare;
        procedure StartMonitoring;
        procedure StopMonitoring;
  public
    constructor Create(AOwner :TCOmponent;
                        AOptions :TNotificationOptions );
    destructor Destroy; override;
    property   OnNotify :TFileNotifyEvent read FOnNotify write FOnNotify;
    property   Options  :TNotificationOptions read FOptions write SetOptions;
    property   Folder   :String read FFolder write SetRootFolder;
    property   AsyncNotify :Boolean read FAsyncNotify write SetAsyncNotify;
  end;

{**********************************************************
                NEW FILE LIST BOX
**********************************************************}
  TFileSelectEvent = procedure (Sender: TObject; Name: string) of object;

  TDirectoryTreeView = class;
  TNewFileListBox = class(TCustomListBox)
  private
        FMask:                  string;
    FFileType:          TFileType;
    FDirectory:         string;
    FDirTree:           TDirectoryTreeView;

    FPrevIconIndex: Integer;
    FIconBitmap:    TBitmap;
        FTempBitmap:    TBitmap;
    FIconWidth:     Integer;
    FIconHeight:    Integer;
    FTextOffset:    Integer;

        FFileNotify:    TFileNotify;
    FNeedRefresh,
    FIsRefreshing:    Boolean;
    FWaitFirstNotify: Boolean;
    FSortedList:          Boolean;
    FNotification:        Boolean;

        FUserOnDblClick:  TNotifyEvent;
    FOnFileSelect:        TFileSelectEvent;
  protected
    procedure   SetFileType(NewFileType: TFileType);
        procedure       SetMask(Value: string);
    procedure   SetDirectory(const NewDirectory: string);
    procedure   SetSortedList(Value: Boolean);
    procedure   SetNotification(Value: Boolean);
        procedure       NewOnDblClick(Sender: TObject);
    procedure   DrawItem(Index: Integer; Rect: TRect;
                                         State: TOwnerDrawState);  override;
        procedure       ChangeFileSystem(Sender :TObject;
                                                         Action :TNotificationOption);
    procedure   FindFolders(DirName: string; DirList: TStrings);
    function    GetFileName: string;
    procedure   Notification(AComponent: TComponent;
                                        Operation: TOperation); override;
  public
        constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
        procedure   FreshList(NewDir: string);
    property    IsRefreshing: Boolean read FIsRefreshing;
    procedure   AddFolders(Value: Boolean);
    property    FileName: string read GetFileName;
    property    DirTree: TDirectoryTreeView read FDirTree write FDirTree;
  published
    property Align;
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property IntegralHeight;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Style;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick: TNotifyEvent
                                        read FUserOnDblClick write FUserOnDblClick;
    property OnFileSelect: TFileSelectEvent
                                        read FOnFileSelect write FOnFileSelect;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property MultiSelect;
    property Directory: string read FDirectory write SetDirectory;
        property Mask: string read Fmask write SetMask;
    property FileType: TFileType read FFileType write SetFileType;
    property Sorted: Boolean
                                        read FSortedList write SetSortedList
                                        default True;
    property Notify: Boolean
                                        read FNotification write SetNotification
                                        default False;
  end;

{**********************************************************
                TCOMPONENT HACK
**********************************************************}
  TComponent_ = class(TPersistent)
  private
    FOwner: TComponent;
    FName: TComponentName;
    FTag: Longint;
    FComponents: TList;
    FFreeNotifies: TList;
    FDesignInfo: Longint;
    FComponentState: TComponentState;
    ///
  end;


{**********************************************************
                DIRECTORY TREE VIEW
**********************************************************}
  TDirectoryTreeView = class(TCustomTreeView)
  protected
    FDirectory:         string;
    FDirLabel:          TLabel;
    FFileListBox:       TNewFileListBox;
    FNotifyList:    TList;
    FNotification:      Boolean;

        FNetBrowse:     Boolean;
        FSorted:                Boolean;
    FNeedRefresh:   Boolean;

    FIsEditingNew:  Boolean;
    FIsSettingPath: Boolean;
    FIsAddingTree:  Boolean;

    FOnEditCancel:    TTVChangedEvent;
    FOnNewEditCancel: TTVNewEditCancelEvent;
        FUserOnEdited:    TTVEditedEvent;
        FUserOnCollapsed: TTVExpandedEvent;
    FUserOnChange:        TTVChangedEvent;

        FOnBeginCreating: TTVBeginCreatingEvent;
        FOnEndCreating:   TTVEndCreatingEvent;

    procedure SetDirLabel(Value: TLabel);
    procedure SetDirLabelCaption;
    procedure SetFileListBox(Value: TNewFileListBox);
    procedure SetDirectory(const NewDirectory: string);
    procedure SetSorted(Value: Boolean);
    procedure SetNetBrowse(Value: Boolean);
    procedure SetNotification(Value: Boolean);
    procedure Edit(const Item: TTVItem); override;
        function  ExtractRootDir(DiskName: string): string;
        function  GetVolumeNodeName(Path: string): string;
        function  GetNodePath(Node: TTreeNode): string;
        function  GetDiskParentNode(Node: TTreeNode): TTreeNode;
    function  CanEdit(Node: TTreeNode): Boolean; override;
    function  CanChange(Node: TTreeNode): Boolean; override;
    function  CanExpand(Node: TTreeNode): Boolean; override;
    function  AllowCreation: Boolean;
        procedure NewOnEdited(Sender: TObject; Node: TTreeNode;
                                                var S: string);
        procedure NewOnCollapsed(Sender: TObject; Node: TTreeNode);
        procedure NewOnChange(Sender: TObject; Node: TTreeNode);
    procedure NewOnDeletion(Sender: TObject; Node: TTreeNode);
        procedure ChangeFileSystem(Sender :TObject;
                                                         Action :TNotificationOption);
        procedure AddDummyChild(Node: TTreeNode);
        procedure DeleteDummyChild(Node: TTreeNode);
        property  CanCreateNode: Boolean read AllowCreation;
        procedure  DeleteNetworkNode(Node: TTreeNode);
        function   GetNetworkRoot: TTreeNode;
        function   GetNetHood(Node: TTreeNode): Boolean;
        function   IsNetworkNode(Node: TTreeNode): Boolean;
    function   IsNetworkEntity(Node: TTreeNode;
                               EntityType: DWORD): Boolean;
        function   IsNetworkContainer(Node: TTreeNode): Boolean;
        function   IsNetworkServer(Node: TTreeNode): Boolean;
        function   IsNetworkResource(Node: TTreeNode): Boolean;
    procedure  AddNetworkRoot(PIDL: PItemIDList);
        procedure  AddSpecialIcons;
    function   NewChildAndEdit(Node: TTreeNode; const S: String)
                                                : TTreeNode;
  public
        constructor Create(AOwner: TComponent); override;
    property    IsEditingNew: Boolean read FIsEditingNew
                                                           write FIsEditingNew;
        destructor  Destroy; override;
        function    AddDirectories(Node: TTreeNode): Boolean;
        procedure   ExpandPath(Path: string; ExpandNode: Boolean);
    procedure   Notification(AComponent: TComponent;
                                        Operation: TOperation); override;
    function    CreateDirectory: Boolean;
  published
    property BorderStyle;
    property Indent;
    property Align;
    property Enabled;
    property Font;
    property Color;
    property ParentColor default False;
    property ParentCtl3D;
    property Ctl3D;
    property TabOrder;
    property TabStop default True;
    property Visible;

    property OnEditing;
    property OnExpanding;
    property OnExpanded;
    property OnCollapsing;
    property OnCompare;
    property OnChanging;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnClick;
    property OnEnter;
    property OnExit;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property PopupMenu;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;

    property DirLabel: TLabel read FDirLabel write SetDirLabel;
    property FileBox: TNewFileListBox
                                        read FFileListBox write SetFileListBox;
    property Sorted: Boolean read FSorted write SetSorted default True;
    property NetBrowse: Boolean read FNetBrowse write SetNetBrowse
                                        default False;
    property Notify: Boolean
                                        read FNotification write SetNotification
                                        default False;
        property Directory: string read FDirectory write SetDirectory;
    property OnEditCancel: TTVChangedEvent
                                        read FOnEditCancel write FOnEditCancel;
    property OnNewEditCancel: TTVNewEditCancelEvent
                                        read FOnNewEditCancel write FOnNewEditCancel;
    property OnEdited: TTVEditedEvent
                                        read FUserOnEdited write FUserOnEdited;
    property OnCollapsed: TTVExpandedEvent
                                        read FUserOnCollapsed write FUserOnCollapsed;
    property OnChange: TTVChangedEvent
                                        read FUserOnChange write FUserOnChange;
        property OnBeginCreating: TTVBeginCreatingEvent
                                read FOnBeginCreating write FOnBeginCreating;
        property OnEndCreating:   TTVEndCreatingEvent
                                read FOnEndCreating write FOnEndCreating;
  end;


//############################################################
function HasSubdirs(DirName: string): Boolean;
function FindNextDirectory(var F: TSearchRec): Boolean;
function FindFirstDirectory(const Path: string;
                                                     var F: TSearchRec): Boolean;
function kpGetImageIndex(Filename: String): Integer;

procedure Register;

//############################################################
implementation

{$IFDEF VER90}
const
  shell32 = 'shell32.dll';

function SHGetSpecialFolderLocation;    external shell32 name
                 'SHGetSpecialFolderLocation';
{$ENDIF}

{$R DIRTREE.RES}

const
  DefaultMask = '*.*';
  DummyChildName = '.';
  MyComputerName = ' ';
  NetworkName = ' ';

var
        SysSmallImageList: TImageList;
        TreeViewImageList:   TImageList;

        MyComputerIcon:          Integer;
    FolderIconIndex:     Integer;
    OpenFolderIconIndex: Integer;
    ServerIconIndex:     Integer;
    ServerNoAccessIconIndex: Integer;
    NetworkIconIndex:    Integer;
    NethoodIconIndex:    Integer;
    DomainIconIndex:     Integer;

{**********************************************************
                CREATE
**********************************************************}
constructor TNewFileListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  Width := 145;
{  IntegralHeight := True;}
  FFileType := [ftNormal]; { show only normal files by default }
  FMask := DefaultMask;
  MultiSelect := False;
  Style := lbOwnerDrawFixed;

  FNeedRefresh  := False;
  FIsRefreshing := False;
  FSortedList := True;
  FNotification := False;
  Directory := '';

  FPrevIconIndex := - 1;
  FIconBitmap := TBitmap.Create;
  FTempBitmap := TBitmap.Create;
  FIconWidth  := - 1;
  FIconHeight := - 1;
  FTextOffset := - 1;

  inherited OnDblClick := NewOnDblClick;
end;

{**********************************************************
                DESTROY
**********************************************************}
destructor TNewFileListBox.Destroy;
begin
        if FFileNotify <> nil then
       FFileNotify.Free;
    FIconBitmap.Free;
    FTempBitmap.Free;
    inherited;
end;

{**********************************************************
                NEW ON DBL CLICK
**********************************************************}
procedure TNewFileListBox.NewOnDblClick(Sender: TObject);
var Node, Child: TTreeNode;
        ItemName: string;
//----------------------------------------------------------
        function FolderSelected: Boolean;
    begin
        Result := False;
        if ItemIndex < 0 then Exit;
        Result :=
                (integer(Items.Objects[ItemIndex]) = FolderIconIndex);
    end;
//----------------------------------------------------------
begin
        if FolderSelected then begin
        if Assigned(DirTree) then begin
                Node := DirTree.Selected;
                if Node <> nil then begin
                        ItemName := Items[ItemIndex];
                if not Node.Expanded then Node.Expand(False);
                Child := Node.GetFirstChild;
                while Child <> nil do
                        if Child.Text = ItemName then begin
                                DirTree.Selected := Child;
                                break;
                        end
                        else Child := Node.GetNextChild(Child);
                end;
        end;
    end
    else
    if (ItemIndex >= 0) and Assigned(FOnFileSelect) then
        OnFileSelect(Sender, Items[ItemIndex]);

    if Assigned(FUserOnDblClick) then FUserOnDblClick(Sender);
end;

{**********************************************************
                NOTIFICATION
**********************************************************}
procedure TNewFileListBox.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then begin
    if AComponent = FDirTree then FDirTree := nil;
  end;
end;

{**********************************************************
                GET FILE NAME
**********************************************************}
function TNewFileListBox.GetFileName;
begin
     if ItemIndex >= 0 then
          Result := Items[ItemIndex]
     else Result := '';
end;

{**********************************************************
                CHANGE FILE SYSTEM
**********************************************************}
procedure TNewFileListBox.ChangeFileSystem(
                                Sender :TObject;Action :TNotificationOption);
begin
    if Parent <> nil then begin
           if not FWaitFirstNotify then
                  if not FIsRefreshing then
                FreshList(FDirectory)
          else FNeedRefresh := True;
       FWaitFirstNotify := False;
    end;
end;

{**********************************************************
                FIND FOLDERS
**********************************************************}
procedure TNewFileListBox.FindFolders(DirName: string; DirList: TStrings);
var
  DirFound: Boolean;
  F : TSearchRec;
begin
    DirList.Clear;
        DirFound := FindFirstDirectory(DirName+'*.*', F);
        while DirFound do begin
        if F.Name[1]<>'.' then
               DirList.AddObject(F.Name,TObject(FolderIconIndex));
        Application.ProcessMessages;
        DirFound := FindNextDirectory(F);
    end;
end;

{**********************************************************
                ADD/REMOVE FOLDERS
**********************************************************}
procedure TNewFileListBox.AddFolders(Value: Boolean);
var IconIndex: integer;
        TempList, OldList: TStringList;
begin
        with Items do begin
          if not Value then
        while Count>0 do begin
                        IconIndex := integer(Objects[0]);
                        if IconIndex<>FolderIconIndex then
                        break;
                Delete(0);
        end
      else begin
                TempList := TStringList.Create;
                OldList := TStringList.Create;
        try
                OldList.Assign(Items);
                TempList.Sorted := Sorted;
            FindFolders(FDirectory,TempList);
                TempList.Sorted := False;
            TempList.AddStrings(OldList);
            Items.Assign(TempList);
        finally
                TempList.Free;
                OldList.Free;
        end;
      end;
    end;
end;

{**********************************************************
                FRESH FILE LIST
**********************************************************}
procedure TNewFileListBox.FreshList(NewDir: string);
const
   Attributes: array[TFileAttr] of Word =
        (faReadOnly, faHidden, faSysFile,
     faVolumeID, faDirectory, faArchive, 0);
var
  AttrIndex: TFileAttr;
  Attr : integer;
  Divider: integer;
  F : TSearchRec;
  DirList, FileList: TStringList;
  MaskList, Mask1: string;
  FileFound:    Boolean;
begin

    ItemHeight := Font.Size * 2;
    Items.Clear;

    if Length(NewDir)=0 then Exit;

    FileList := TStringList.Create;
    FileList.Sorted := Sorted;
    DirList := TStringList.Create;
    DirList.Sorted := Sorted;

        Screen.Cursor := crHourGlass;
    SetCapture(Handle);
    FIsRefreshing := True;

  try
   repeat
     FileList.Clear;
     FNeedRefresh  := False;

         if ftDirectory in FFileType then
                FindFolders(NewDir,DirList);

     if FNeedRefresh then continue;

     // Set attribute flags based on values in FileType
     Attr := 0;
     for AttrIndex := ftReadOnly to ftArchive do
      if (AttrIndex <> ftDirectory)  and
         (AttrIndex in FileType) then
        Attr := Attr or Attributes[AttrIndex];

     MaskList := FMask;
     while Length(MaskList)>0 do begin
        Divider := Pos(';',MaskList);
        if Divider > 0 then begin
                Mask1 := Copy(MaskList,1,Divider-1);
            MaskList := Copy(MaskList,Divider+1,999);
        end
        else begin
                Mask1 := MaskList;
            MaskList := '';
        end;
                FileFound := FindFirst(NewDir+Mask1, Attr, F) = 0;
        while FileFound do begin
                        F.Name := AnsiLowerCase(F.Name);
            if FileList.IndexOf(F.Name) < 0 then
                        FileList.AddObject(
                        F.Name, TObject(kpGetImageIndex(NewDir+F.Name)));
                Application.ProcessMessages;
                        FileFound := FindNext(F) = 0;
                if not FileFound then FindClose(F);
                end;
     end;

   until not FNeedRefresh;

  finally
    inherited Sorted := False;

    with Items do begin
         BeginUpdate;
         AddStrings(DirList);
         AddStrings(FileList);
         DirList.Free;
         FileList.Free;
         EndUpdate;
    end;

        ReleaseCapture;
    Screen.Cursor := crDefault;
    FIsRefreshing := False;
  end;
end;

{**********************************************************
                SET MASK
**********************************************************}
procedure TNewFileListBox.SetMask(Value: string);
begin
        if AnsiCompareText(Fmask,Value)<>0 then begin
        FMask := Value;
        FreshList(FDirectory);
    end;
end;

{**********************************************************
                SET SORTED LIST
**********************************************************}
procedure TNewFileListBox.SetSortedList(Value: Boolean);
begin
        if FSortedList<>Value then begin
        FSortedList := Value;
        FreshList(FDirectory);
    end;
end;

{**********************************************************
                SET NOTIFICATION
**********************************************************}
procedure TNewFileListBox.SetNotification(Value: Boolean);
var NotificationOptions: TNotificationOptions;
begin
        if Value <> FNotification then begin
       FNotification := Value;

       if FNotification then begin
          NotificationOptions := [fnoFile];
          if ftDirectory in FFileType then
             NotificationOptions := [fnoFile,fnoFolder]
          else
             NotificationOptions := [fnoFile];

          FFileNotify := TFileNotify.Create(Self,NotificationOptions);
          with FFileNotify do begin
                    AsyncNotify := False;
            Folder := FDirectory;
                    OnNotify := ChangeFileSystem;
            Prepare;
          end;
          if not (csDesigning in ComponentState) then
                    FFileNotify.StartMonitoring;
       end
       else begin
          FFileNotify.Free;
          FFileNotify := nil;
       end;
    end;
end;

{**********************************************************
                SET FILE TYPE
**********************************************************}
procedure   TNewFileListBox.SetFileType(NewFileType: TFileType);
var     AddFoldersOnly,
        HideFoldersOnly: Boolean;
begin
        if NewFileType <> FFileType then begin

        if  (FFileNotify<>nil) and
                ((ftDirectory in NewFileType) <>
                         (fnoFolder in FFileNotify.Options)) then
            if (ftDirectory in NewFileType) then
                 FFileNotify.Options := [fnoFile,fnoFolder]
            else FFileNotify.Options := [fnoFile];

        AddFoldersOnly := (NewFileType-FFileType)= [ftDirectory];
        HideFoldersOnly := (FFileType-NewFileType) = [ftDirectory];
        FFileType := NewFileType;

        if AddFoldersOnly or HideFoldersOnly then
                AddFolders(AddFoldersOnly)
        else
                FreshList(FDirectory);
    end;
end;

{**********************************************************
                SET DIRECTORY
**********************************************************}
procedure TNewFileListBox.SetDirectory(const NewDirectory: string);
begin

  if NewDirectory='...' then begin
        FDirectory := NewDirectory;
    Clear;
        Exit;
  end;

{$IFNDEF VER90}
  if (AnsiCompareFileName(NewDirectory, Directory) = 0)
{$ELSE}
  if (AnsiCompareStr(AnsiLowerCase(NewDirectory),
                     AnsiLowerCase(Directory)) = 0)
{$ENDIF}
         then Exit;

  FDirectory := NewDirectory;
  FreshList(FDirectory);

  if FFileNotify<>nil then begin
        if FFileNotify.Folder<>'' then
        FWaitFirstNotify := True;
        FFileNotify.Folder := FDirectory;
  end;
end;

{**********************************************************
                DRAW ITEM
**********************************************************}
procedure TNewFileListBox.DrawItem(Index: Integer; Rect: TRect;
                                State: TOwnerDrawState);
var     IconIndex: integer;
    rct:         TRect;
begin
        with Canvas do begin
        FillRect(Rect);
        IconIndex := integer(Items.Objects[Index]);
        if IconIndex <> FPrevIconIndex then begin
            SysSmallImageList.GetBitmap(IconIndex,FTempBitmap);
            if FIconWidth < 0 then begin
               FIconWidth  := MulDiv(FTempBitmap.Width,ItemHeight,16);
               FIconHeight := MulDiv(FTempBitmap.Height,ItemHeight,16);
               FTextOffset := FIconWidth + 8;
               FIconBitmap.Width := FIconWidth;
               FIconBitmap.Height := FIconHeight;
            end;
            rct := Classes.Rect(0,0,FIconWidth-1,
                                FIconHeight-1 );
            FIconBitmap.Canvas.StretchDraw(rct,FTempBitmap);
        end;

        Draw(Rect.Left+2,Rect.Top,FIconBitmap);
        TextOut(Rect.Left+FTextOffset,Rect.Top,Items[Index]);

        FPrevIconIndex := IconIndex;
        end;
end;

{**********************************************************
                HAS SUB DIRS
**********************************************************}
function HasSubdirs(DirName: string): Boolean;
var    F : TSearchRec;
begin
        Result := FindFirstDirectory(DirName+'*.*', F);
    if Result then FindClose(F);
end;

{**********************************************************
                FIND NEXT DIRECTORY
**********************************************************}
function FindNextDirectory(var F: TSearchRec):Boolean;
begin
        Result := False;
    while not Result do begin
        if FindNext(F)<>0 then begin
                FindClose(F);
            Exit;
        end;
        if ((faDirectory and f.Attr)<>0) and
              (F.Name[1] <> '.') then begin
                Result := True;
            Exit;
        end;
    end;
end;

{**********************************************************
                FIND FIRST DIRECTORY
**********************************************************}
function FindFirstDirectory(const Path: string;
                                                        var F: TSearchRec): Boolean;
begin
        Result := FindFirst(Path, faDirectory, F) = 0;
        if Result then begin
        if ((faDirectory and f.Attr)<>0) and
           (F.Name[1] <> '.') then
           Exit;
        Result := FindNextDirectory(F);
    end;
end;

{**********************************************************
                GET VOLUME NODE NAME
**********************************************************}
function TDirectoryTreeView.GetVolumeNodeName(Path: string): string;
var
        VolumeLabel, FileSystemName: array [0..100] of char;
    VolumeSerialNumber,
    MaximumComponentLength, FileSystemFlags: DWORD;
begin
    Result := ExtractRootDir(Path);
        if GetVolumeInformation(
                PChar(Path+'\'), VolumeLabel, 100,
            @VolumeSerialNumber,
            MaximumComponentLength, FileSystemFlags,
            FileSystemName, 100 ) then begin
            if StrLen(VolumeLabel) > 0 then
                Result := VolumeLabel[0]+
                    AnsiLowerCase(StrPas(VolumeLabel+1))+
                    ' ('+Result+')';
    end;
end;

{**********************************************************
                ADD SPECIAL ICONS
**********************************************************}
procedure TDirectoryTreeView.AddSpecialIcons;
var     Icon: TIcon;
{**********************************************************}
        procedure AddIconToList(Name: string; var IconIndex: integer);
    begin
        Icon.Handle := LoadIcon(hInstance,PChar(Name));
        if Icon.Handle<>0 then begin
                TreeViewImageList.AddIcon(Icon);
            IconIndex := TreeViewImageList.Count-1;
        end;
    end;

{**********************************************************}
begin
    if OpenFolderIconIndex >= 0 then Exit;

    TreeViewImageList.Assign(SysSmallImageList);

    OpenFolderIconIndex := FolderIconIndex;
        NetworkIconIndex := 0;
        DomainIconIndex  := 0;
        ServerIconIndex  := 0;
    ServerNoAccessIconIndex := 0;
    
    Icon := TIcon.Create;
    try
        try
          AddIconToList('ZZZ_OFOLDER',OpenFolderIconIndex);
//          AddIconToList('ZZZ_NETHOOD',NethoodIconIndex);
          AddIconToList('ZZZ_NETICON',NetworkIconIndex);
          AddIconToList('ZZZ_DOMNICON',DomainIconIndex);
          AddIconToList('ZZZ_SERVICON',ServerIconIndex);
          AddIconToList('ZZZ_SERVNOICON',ServerNoAccessIconIndex);
        finally
          Icon.Free;
        end;
        except
        end;
end;

{**********************************************************
                GET NETHOOD
**********************************************************}
function TDirectoryTreeView.GetNetHood(Node: TTreeNode): Boolean;
type
  PNetResourceArray = ^TNetResourceArray;
  TNetResourceArray =
                        array[0..MaxInt div SizeOf(TNetResource)-1]
               of TNetResource;
var
  I, BufSize, NetResult, Count: Integer;
  Size: DWORD;
  NetHandle: THandle;
  NetResources: PNetResourceArray;
  Ptr: PNetResource;
  S: String;
  NewNode : TTreeNode;
  NetNodeList: TStringList;

begin

  Result := False;
  BufSize := 50 * SizeOf(TNetResource);
  GetMem(NetResources, BufSize);
  NetNodeList := TStringList.Create;
  NetNodeList.Sorted := Sorted;

  WNetOpenEnum(RESOURCE_GLOBALNET, RESOURCETYPE_DISK,
               RESOURCEUSAGE_CONNECTABLE or RESOURCEUSAGE_CONTAINER,
               PNetResource(Node.Data),
               NetHandle);

//-----------------------------------------------------------------

  try
     while True do begin
          Count := - 1;
          Size := BufSize;
          NetResult := WNetEnumResource(NetHandle, DWORD(Count), NetResources, Size);
          if NetResult = ERROR_MORE_DATA then begin
                BufSize := Size;
                ReallocMem(NetResources, BufSize);
                Continue;
          end;
          if NetResult <> NO_ERROR then Break;

          Result := Count > 0;

          for I := 0 to Count-1 do begin
            New(Ptr);
            Move(NetResources[I], Ptr^, SizeOf(TNetResource));
            if NetResources[I].lpLocalName <> nil then
               Ptr^.lpLocalName := StrNew(NetResources[I].lpLocalName);
            if NetResources[I].lpComment <> nil then
               Ptr^.lpComment := StrNew(NetResources[I].lpComment);
            if NetResources[I].lpRemoteName <> nil then
               Ptr^.lpRemoteName := StrNew(NetResources[I].lpRemoteName);

            if NetResources[I].lpRemoteName[1] = '\' then
                 S := ExtractFileName(NetResources[I].lpRemoteName)
            else S := NetResources[I].lpRemoteName;
            if NetResources[I].lpProvider <> nil then
               Ptr^.lpProvider := StrNew(NetResources[I].lpProvider);

            NetNodeList.AddObject(S,TObject(Ptr));
         end;
     end;
//------------------------------------------------------------------
  finally
     WNetCloseEnum(NetHandle);
     FreeMem(NetResources, BufSize);

     if IsNetworkServer(Node) then
        with Node do begin
             if Result then
                ImageIndex := ServerIconIndex
             else ImageIndex := ServerNoAccessIconIndex;
                  SelectedIndex := ImageIndex
        end;

     for i:=0 to NetNodeList.Count-1 do begin
         Ptr := PNetResource(NetNodeList.Objects[i]);
         NewNode := Node.Owner.AddChildObject(Node,
                    NetNodeList[i], Ptr);
         with NewNode do begin
                HasChildren := True;
            SelectedIndex := 0;
                case Ptr^.dwDisplayType of
                        RESOURCEDISPLAYTYPE_NETWORK:
                        ImageIndex := NetworkIconIndex;
                RESOURCEDISPLAYTYPE_DOMAIN:
                        ImageIndex := DomainIconIndex;
                RESOURCEDISPLAYTYPE_SERVER:
                        ImageIndex := ServerIconIndex;
                RESOURCEDISPLAYTYPE_SHARE:
                                begin
                        ImageIndex := FolderIconIndex;
                                        SelectedIndex := OpenFolderIconIndex;
                        end;
                end;
                if SelectedIndex = 0 then SelectedIndex := ImageIndex;
                end;
                    //     
                    // 
//            if Ptr^.dwDisplayType <> RESOURCEDISPLAYTYPE_SHARE then
//                  GetNetHood(NewNode);
     end;
     NetNodeList.Free;
  end;
end;

{**********************************************************
                IS NETWORK NODE
**********************************************************}
function TDirectoryTreeView.IsNetworkNode(Node: TTreeNode): Boolean;
var     ParentNode: TTreeNode;
begin
        ParentNode := GetDiskParentNode(Node);
    Result := False;
    if FNetBrowse {and (Node.Text<>DummyChildName) }then
       if ParentNode.Parent = nil then
          Result := (ParentNode.Text <> MyComputerName)
       else
          Result := (ParentNode.Parent.Text = NetworkName);
end;

{**********************************************************
                IS NETWORK TYPE
**********************************************************}
function TDirectoryTreeView.IsNetworkEntity(Node: TTreeNode;
                 EntityType: DWORD): Boolean;
var  NetResource: PNetResource;
begin
        Result := False;
    if IsNetworkNode(Node) then begin
        NetResource := PNetResource(Node.Data);
        if NetResource<>nil then
           Result := (NetResource^.dwDisplayType = EntityType);
    end;
end;

{**********************************************************
                IS NETWORK CONTAINER
**********************************************************}
function TDirectoryTreeView.IsNetworkContainer(Node: TTreeNode): Boolean;
var  NetResource: PNetResource;
begin
    NetResource := PNetResource(Node.Data);
        Result := IsNetworkNode(Node)
              and (NetResource<>nil)
              and not IsNetworkResource(Node);
end;

{**********************************************************
                IS NETWORK RESOURCE
**********************************************************}
function TDirectoryTreeView.IsNetworkResource(Node: TTreeNode): Boolean;
begin
        Result := IsNetworkEntity(Node,RESOURCEDISPLAYTYPE_SHARE);
end;

{**********************************************************
                IS NETWORK SERVER
**********************************************************}
function TDirectoryTreeView.IsNetworkServer(Node: TTreeNode): Boolean;
begin
        Result := IsNetworkEntity(Node,RESOURCEDISPLAYTYPE_SERVER);
end;

{**********************************************************
                DELETE NETWORK NODE
**********************************************************}
procedure  TDirectoryTreeView.DeleteNetworkNode(Node: TTreeNode);
var  Child, NodeToDelete: TTreeNode;
begin
        if IsNetworkNode(Node) then begin
                Child := Node.GetLastChild;
        while Child<>nil do begin
            NodeToDelete := Child;
                Child := Node.GetPrevChild(Child);
                DeleteNetworkNode(NodeToDelete);
        end;
        Items.Delete(Node);
        end;
end;

{**********************************************************
                SET NOTIFICATION
**********************************************************}
procedure TDirectoryTreeView.SetNotification(Value: Boolean);
var  DiskFileNotify: TFileNotify;
     DiskNode: TTreeNode;
begin
    if Value <> FNotification then begin

           FNotification := Value;
       if csDesigning in ComponentState then Exit;

       DiskNode := Items[0].GetFirstChild;
       while DiskNode <> nil do begin
             DiskFileNotify := TFileNotify(DiskNode.Data);
             if (DiskFileNotify <> nil) and DiskNode.Expanded then
                if FNotification then
                          DiskFileNotify.StartMonitoring
                else  DiskFileNotify.StopMonitoring;
             DiskNode := Items[0].GetNextChild(DiskNode);
       end;
    end;
end;

{**********************************************************
                SET NET BROWSE
**********************************************************}
procedure TDirectoryTreeView.SetNetBrowse(Value: Boolean);
var NetRoot:  TTreeNode;
    PIDL:  PItemIDList;
begin
        if FNetBrowse <> Value then begin
        if not FNetBrowse then begin
                                        //  " "
                if SHGetSpecialFolderLocation(Handle,
                        CSIDL_NETWORK, PIDL) = NOERROR then
                        AddNetworkRoot(PIDL);
        end                     
        else begin
                NetRoot := GetNetworkRoot;
                if NetRoot <> nil then DeleteNetworkNode(NetRoot);
        end;
                FNetBrowse := Value;
    end;
end;

{**********************************************************
                GET NETWORK ROOT
**********************************************************}
function TDirectoryTreeView.GetNetworkRoot: TTreeNode;
var     Index: integer;
begin
        Result := nil;
                //   Items[0] = " "
        for Index:=1 to Items.Count-1 do
        if Items[Index].Parent = nil then begin //  
                Result := Items[Index];
            break;
        end;
end;

{**********************************************************
                ADD NETWORK ROOT
**********************************************************}
procedure  TDirectoryTreeView.AddNetworkRoot(PIDL: PItemIDList);
var
        NewNode: TTreeNode;
    NetResource: PNetResource;
        NetHandle: THandle;
begin
        New(NetResource);
        with NetResource^ do begin
        dwScope := RESOURCE_GLOBALNET;
                dwType := RESOURCETYPE_ANY;
                dwDisplayType := RESOURCEDISPLAYTYPE_SERVER;
                dwUsage := RESOURCEUSAGE_CONNECTABLE;
                lpLocalName := Nil;
                lpRemoteName := Nil;
                lpComment := StrNew('');
                lpProvider := StrNew('');
    end;

                // ,   

    if WNetOpenEnum(RESOURCE_GLOBALNET,
                         RESOURCETYPE_DISK,
                 RESOURCEUSAGE_CONNECTABLE or RESOURCEUSAGE_CONTAINER,
                 PNetResource(NetResource),
                 NetHandle) <> NO_ERROR then begin
                with NetResource^ do begin
                        StrDispose(lpComment);
                        StrDispose(lpProvider);
        end;
        Dispose(NetResource);
        Exit;
    end;

    WNetCloseEnum(NetHandle);

                //  ,     

    NewNode := Items.AddChildObject(nil,
                                NetworkName, NetResource);
    with NewNode do begin
        HasChildren := True;
        ImageIndex := NethoodIconIndex;
        SelectedIndex := ImageIndex;
    end;

    GetNetHood(NewNode);  //    
                                          //   

end;

{**********************************************************
                CREATE
**********************************************************}
constructor TDirectoryTreeView.Create(AOwner: TComponent);
var DirName: string;
    DriveNo: Byte;
    DriveType: integer;
        NewNode: TTreeNode;
    FileNotify: TFileNotify;

        {**********************************************************
                DRIVE EXISTS
        **********************************************************}
        function DriveExists(Drive:Byte):Boolean;
        var
                LogDrives: set of 0..25;
        begin
                integer(LogDrives):=GetLogicalDrives;
                Result:=Drive in LogDrives
        end;

{**********************************************************}
begin
        inherited Create(AOwner);

    FIsSettingPath := False;

    inherited OnEdited    := NewOnEdited;
    inherited OnCollapsed := NewOnCollapsed;
    inherited OnChange    := NewOnChange;
    inherited OnDeletion  := NewOnDeletion;

                //  " "

        Self.Parent := TWinControl(Self.Owner);
    Items.Add(nil,MyComputerName);
    with TopItem do begin
        ImageIndex := MyComputerIcon;
        SelectedIndex := ImageIndex;
    end;

    FNotifyList := TList.Create;

                //     

        for DriveNo := 0 to 25 do
        if DriveExists(DriveNo) then begin
                DirName := Char(DriveNo+$41)+':';
            DriveType := GetDriveType(PChar(DirName));

            FileNotify := nil;
            if (DriveType = DRIVE_FIXED) or
               (DriveType = DRIVE_RAMDISK) then begin
               FileNotify := TFileNotify.Create(Self,
                              [fnoFolder,fnoWatchSubTree]);
               FNotifyList.Add(FileNotify);
               with FileNotify do begin
                         AsyncNotify := False;
                         OnNotify := ChangeFileSystem;
                 Folder := DirName+'\';
                 Prepare;
               end;
               if FNotification and
                  not (csDesigning in ComponentState) then
                      FileNotify.StartMonitoring
               else
                      FileNotify.StopMonitoring;
            end;

                        NewNode:=Items.AddChildObject(Items[0],DirName,FileNotify);

                        if (DriveNo>1) then
                NewNode.Text := GetVolumeNodeName(NewNode.Text);

            AddDummyChild(NewNode);
            with NewNode do begin
                ImageIndex:=kpGetImageIndex(DirName+'\');
                SelectedIndex:=ImageIndex;
                        end;
        end;

        Images := TreeViewImageList;

    AddSpecialIcons;

    FNeedRefresh := False;
    FSorted := True;
    FNetBrowse := False;
    FNotification := False;
end;

{**********************************************************
       IS NODE EXPANDED
**********************************************************}
function IsNodeExpanded(Node: TTreeNode): Boolean;
var ChildNode : TTreeNode;
begin
        ChildNode := Node.GetFirstChild;
    Result := (ChildNode<>nil) and
                  (ChildNode.Text<>DummyChildName);
end;

{**********************************************************
                CHANGE FILE SYSTEM
**********************************************************}
procedure TDirectoryTreeView.ChangeFileSystem(
                                Sender :TObject;
                Action :TNotificationOption);
var
        OldDir: string;
     {**********************************************************
                        IS DRIVE EXPANDED
     **********************************************************}
     function IsDriveExpanded(NotifyFrom: TFileNotify): Boolean;
     var
        DriveNode: TTreeNode;
     begin
          Result := False;
          DriveNode := Items[0].GetFirstChild;
          while DriveNode <> nil do begin
                if TFileNotify(DriveNode.Data)=NotifyFrom
                   then begin
                   Result := IsNodeExpanded(DriveNode);
                   Exit;
                end;
                DriveNode := Items[0].GetNextChild(DriveNode);
          end;
     end;
//-------------------------------------------------------------
begin
    if IsDriveExpanded(TFileNotify(Sender)) then begin
           if not FIsSettingPath then begin
         OldDir := FDirectory;
         FDirectory := ' ';
         ExpandPath(OldDir,IsNodeExpanded(Selected));
       end
       else
         FNeedRefresh := True;
        end;
end;

{**********************************************************
                DESTROY
**********************************************************}
destructor TDirectoryTreeView.Destroy;
begin
    inherited Destroy;
    if FNotifyList <> nil then
        FNotifyList.Free;
end;

{**********************************************************
                SET DIR LABEL
**********************************************************}
procedure TDirectoryTreeView.SetDirLabel (Value: TLabel);
begin
  FDirLabel := Value;
  if Value <> nil then Value.FreeNotification(Self);
  SetDirLabelCaption;
end;

{**********************************************************
                SET DIR LABEL CAPTION
**********************************************************}
procedure TDirectoryTreeView.SetDirLabelCaption;
var
  DirWidth: Integer;
begin
  if FDirLabel <> nil then
  begin
    DirWidth := Width;
    if not FDirLabel.AutoSize then DirWidth := FDirLabel.Width;
    if FDirectory = '...' then
          FDirLabel.Caption := ''
    else  FDirLabel.Caption :=
            MinimizeName(Directory, FDirLabel.Canvas, DirWidth);
  end;
end;

{**********************************************************
                SET FILE LIST BOX
**********************************************************}
procedure TDirectoryTreeView.SetFileListBox(Value: TNewFileListBox);
begin
  FFileListBox := Value;
  if Value <> nil then begin
        Value.FreeNotification(Self);
    Value.DirTree := Self;
    if Selected <> nil {not csDesigning in ComponentState} then
        NewOnChange(Self,Selected);
  end;
end;

{**********************************************************
                SET SORTED
**********************************************************}
procedure TDirectoryTreeView.SetSorted(Value: Boolean);
begin
  if FSorted<>Value then begin
     FSorted := Value;
     ExpandPath(FDirectory,IsNodeExpanded(Selected)); 
  end;
end;

{**********************************************************
                SET DIRECTORY
**********************************************************}
procedure TDirectoryTreeView.SetDirectory(const NewDirectory: string);
begin
  FIsSettingPath := True;
  try
        if Length (NewDirectory) = 0 then Exit;

{$IFNDEF VER90}
   if (AnsiCompareFileName(NewDirectory, FDirectory) = 0)
{$ELSE}
  if (AnsiCompareStr(AnsiLowerCase(NewDirectory),
                     AnsiLowerCase(FDirectory)) = 0)
{$ENDIF}
         then Exit;
        repeat
        FNeedRefresh := False;
                ExpandPath(NewDirectory,False);
        until not FNeedRefresh;
  finally
        FIsSettingPath := False;
  end;
end;

{**********************************************************
                GET DISK PARENT NODE
**********************************************************}
function TDirectoryTreeView.GetDiskParentNode(Node: TTreeNode): TTreeNode;
begin
        Result := Node;
    while (Result.Parent <> nil)  and
          (Result.Parent.Parent<>nil) do
        Result := Result.Parent;
end;

{**********************************************************
                GET NODE PATH
**********************************************************}
function TDirectoryTreeView.GetNodePath(Node: TTreeNode): string;
var TempNode, ParentDiskNode: TTreeNode;
begin
    Result := '';
        if Node.Level < 1 then begin
        Result := '...';
        Exit;
    end;

        ParentDiskNode := GetDiskParentNode(Node);
    if Node.Level > 1 then      Result := Node.Text+'\';

    TempNode := Node.Parent;
    if IsNetworkNode(Node) then begin
       if IsNetworkContainer(Node) then begin
          Result := '';
          Exit;
       end;
       while not IsNetworkServer(TempNode) do begin
             TempNode := TempNode.Parent;
             if TempNode = nil then begin
                Result := '';
                Exit;
             end;
       end;
       ParentDiskNode := TempNode;
    end;

    TempNode := Node.Parent;
    if TempNode.Parent <> nil then
        while TempNode <> ParentDiskNode do begin
                Result := TempNode.Text+'\'+Result;
                TempNode := TempNode.Parent;
            if IsNetworkServer(TempNode) then break;
        end;

    if not IsNetworkNode(ParentDiskNode) then
         Result := ExtractRootDir(ParentDiskNode.Text)+'\'+Result
    else Result := '\\'+ParentDiskNode.Text+'\'+Result;
end;

{**********************************************************
                ADD DUMY CHILD
**********************************************************}
procedure TDirectoryTreeView.AddDummyChild(Node: TTreeNode);
begin
        Node.Owner.AddChild(Node, DummyChildName);
end;

{**********************************************************
                DELETE DUMMY CHILD
**********************************************************}
procedure TDirectoryTreeView.DeleteDummyChild(Node: TTreeNode);
var DummyNode: TTreeNode;
begin
        DummyNode := Node.GetFirstChild;
    while DummyNode<>nil do begin
        if DummyNode.Text=DummyChildName then
                break;
                DummyNode := Node.GetNextChild(DummyNode);
        end;

    if DummyNode<>nil then
        Node.Owner.Delete(DummyNode);
end;

{**********************************************************
                ADD DIRECTORIES
**********************************************************}
function TDirectoryTreeView.AddDirectories(Node: TTreeNode): Boolean;
var NewNode: TTreeNode;
    i: integer;
    DirFound: Boolean;
    F : TSearchRec;
        DirName: string;
        DirList: TStringList;
begin

        DirName := GetNodePath(Node);
        Result := False;

    if (csDesigning in ComponentState) and
       (IsNetworkNode(Node) or (DirName[1] < 'C')) then Exit;

    Items.BeginUpdate;
    Screen.Cursor := crHourGlass;
    SetCapture(Handle);

    FIsAddingTree := True;

        if IsNetworkNode(Node) and (Node.Data<>nil) and
       (PNetResource(Node.Data)^.dwDisplayType<>
         RESOURCEDISPLAYTYPE_SHARE) then begin
         Result := GetNethood(Node);
    end
    else begin
      DirList := TStringList.Create;
          DirList.Sorted := Sorted;

          DirFound := FindFirstDirectory(DirName+'*.*', F);

          while DirFound do begin
        Result := True;
        DirList.Add(F.Name);
        Application.ProcessMessages;
                DirFound := FindNextDirectory(F);
          end;

      for i:=0 to DirList.Count-1 do begin
        NewNode:=Items.AddChild(Node,DirList[i]);
        with NewNode do begin
            ImageIndex := FolderIconIndex;
            SelectedIndex := OpenFolderIconIndex;
        end;
        if HasSubdirs(DirName+DirList[i]+'\') then
                AddDummyChild(NewNode);
      end;

      DirList.Free;
    end;

    Items.EndUpdate;
    Screen.Cursor := crDefault;
    ReleaseCapture;
    FIsAddingTree := False;

end;

{**********************************************************
        GET IMAGE INDEX
**********************************************************}
function kpGetImageIndex(Filename: String): Integer;
var
  Fileinfo: TSHFileInfo;
begin
  if SHGetFileInfo(PChar(FileName),
        0, Fileinfo, sizeof(TSHFileInfo),
        SHGFI_ICON or SHGFI_SYSICONINDEX ) <> 0 then begin
    Result := Fileinfo.IIcon;
  end
  else
    Result := 0;
end;

{**********************************************************
                EXTRACT ROOT DIR
**********************************************************}
function TDirectoryTreeView.ExtractRootDir(DiskName: string): string;
var Len: integer;
begin
        Result := DiskName;
        Len := Length(DiskName);
    if DiskName[Len]=')' then
        Result := Copy(DiskName,Len-2,2);
end;

{**********************************************************
                NEW ON DELETION
**********************************************************}
procedure TDirectoryTreeView.NewOnDeletion(Sender: TObject;
                                           Node: TTreeNode);
var NetResource: PNetResource;
    FileNotify: TFileNotify;
begin
     if Node.Data <> nil then
        if FNotifyList.IndexOf(Node.Data) < 0 then begin
           NetResource := PNetResource(Node.Data);
           with NetResource^ do begin
                if lpLocalName <> nil then
                   StrDispose(lpLocalName);
                if lpComment <> nil then
                   StrDispose(lpComment);
                if lpRemoteName <> nil then
                   StrDispose(lpRemoteName);
                if lpProvider <> nil then
                   StrDispose(lpProvider);
                end;
           Dispose(NetResource);
        end
        else begin
             FileNotify := TFileNotify(Node.Data);
             FileNotify.Free;
        end;

     Node.Data := nil;

end;

{**********************************************************
                NEW ON EDITED
**********************************************************}
procedure TDirectoryTreeView.NewOnEdited(Sender: TObject;
                        Node: TTreeNode; var S: string);
var Path:       string;
begin
        if Assigned(FUserOnEdited) then
        FUserOnEdited(Sender, Node, S);

//    Exclude(TComponent_(Self).FComponentState, csDesigning);

        if (Node<>nil) and FIsEditingNew then begin
                Node.Text := S;
                Path := GetNodePath(Node);
        if not CreateDir(Path) then begin
                MessageDlg('   '+
                        AnsiUpperCase(Path), mtError,
                [mbOK],0 );
            Items.Delete(Node);
            S := '';
        end
        else Selected := Node;
    end;
end;

{**********************************************************
                EDIT
**********************************************************}
procedure TDirectoryTreeView.Edit(const Item: TTVItem);
var
  S: string;    
  Node: TTreeNode;
  Action: Boolean;
begin
  with Item do begin
    { Get the node }
    if (state and TVIF_PARAM) <> 0 then
      Node := Pointer(lParam)
    else
      Node := Items.GetNode(hItem);

    if pszText = nil then begin
      if FIsEditingNew then begin
        Action := True;
        if Assigned(FOnNewEditCancel) then
          FOnNewEditCancel(Self, Node, Action);
        if Action then
          Node.Destroy
      end
      else
        if Assigned(FOnEditCancel) then
          FOnEditCancel(Self, Node);
    end
    else begin
//      inherited;
                with Item do
        if pszText <> nil then begin
                S := pszText;
                NewOnEdited(Self, Node, S);
                if (Node <> nil) and (S <> '') then Node.Text := S;
        end;
        end;
  end;
  FIsEditingNew := False;
  if Assigned(FOnEndCreating) then
                FOnEndCreating(Self,Node);
end;

{**********************************************************
                CREATE DIRECTORY
**********************************************************}
function TDirectoryTreeView.CreateDirectory: Boolean;
begin
        if Assigned(FOnBeginCreating) then
                FOnBeginCreating(Self,Selected);
//    Include(TComponent_(Self).FComponentState, csDesigning);
        Result := NewChildAndEdit(Selected,'-----')<>nil;
end;

{**********************************************************
                NEW CHILD AND EDIT
**********************************************************}
function TDirectoryTreeView.NewChildAndEdit
                        (Node: TTreeNode; const S: String): TTreeNode;
var RootPath: string;
        NetResource: PNetResource;
begin
        Result := nil;

        if IsNetworkNode(Node) then begin
        NetResource := PNetResource(Node.Data);
        if NetResource^.dwDisplayType <>
                                   RESOURCEDISPLAYTYPE_SHARE then Exit;
        end
    else begin
        RootPath := Copy(GetNodePath(Node),1,2)+'\';
                if (Node = nil) or (Node.Parent = nil) or
        (GetDriveType(PChar(RootPath)) in [0,1,DRIVE_CDROM]) then Exit;
    end;

    SetFocus;

    Items.BeginUpdate;

    if (not Node.Expanded) and
       (Node.GetFirstChild <> nil) then
       Node.Expand(False);

    Result := Items.AddChildFirst(Node,S);

    if not Node.Expanded then Node.Expand(False);

    FIsEditingNew := True;
    with Result do begin
        ImageIndex := FolderIconIndex;
        SelectedIndex := OpenFolderIconIndex;
    end;

    Items.EndUpdate;

    Include(TComponent_(Self).FComponentState, csDesigning);
    Result.EditText;
    Exclude(TComponent_(Self).FComponentState, csDesigning);
    SetFocus;
end;

{**********************************************************
                CAN EXPAND
**********************************************************}
function TDirectoryTreeView.CanExpand(Node: TTreeNode): Boolean;
var
        ParentDiskNode: TTreeNode;
    DiskLetter: string;
    DiskFileNotify: TFileNotify;
begin

    if (Node.Level > 1) and (not FIsSettingPath) and
       (csDesigning in ComponentState) then begin
        Result := False;
        Exit;
    end;

        Result := True;
        with Node do
        if Level>0 then begin
            Items.BeginUpdate;
                ParentDiskNode := GetDiskParentNode(Node);

            if ((csDesigning in ComponentState)
                and (not FIsSettingPath)) or FIsAddingTree then
                 Result := False
            else Result := FIsEditingNew or AddDirectories(Node);

            if (not IsNetworkNode(Node))  then begin
                DiskLetter := ExtractRootDir(ParentDiskNode.Text);
                if (Level=1) and
                        (Result or (DiskLetter[1]>'B')) then
                        ParentDiskNode.Text := GetVolumeNodeName(DiskLetter);
            end;
            if Result then DeleteDummyChild(Node);
            Items.EndUpdate;
        end;

    if Result then
        Result := inherited  CanExpand(Node);

    if Result then
       with Node do begin
//        Selected := Node;
//      FDirectory := GetNodePath(Selected);

           if ImageIndex = FolderIconIndex then
                          ImageIndex := OpenFolderIconIndex;

           if not IsNetworkNode(Node) and  (Level = 1) then begin
              DiskFileNotify := TFileNotify(Data);
              if (DiskFileNotify <> nil) and FNotification then
                 DiskFileNotify.StartMonitoring;
           end;
        end;
end;

{**********************************************************
                EXPAND PATH
**********************************************************}
procedure TDirectoryTreeView.ExpandPath(
                        Path: string; ExpandNode: Boolean);
var DirName:    string;
        NodeToExpand, Node, TargetNode: TTreeNode;
    FormParent: TWinControl;
{**********************************************************}
        function FindChild (Node: TTreeNode; NodeName: string;
                                    Recursive: Boolean): TTreeNode;
    var TempNode: TTreeNode;
                RealName:       string;
        InNetwork:      Boolean;
    begin
        Result := nil;
        InNetwork := IsNetworkNode(Node);
                TempNode := Node.GetFirstChild;
        while TempNode<>nil do begin

                RealName := TempNode.Text;
            if (TempNode.Level=1)  and (not InNetwork)
               then RealName := ExtractRootDir(realName);

{$IFNDEF VER90}
            if 0 = AnsiCompareFileName(RealName,NodeName)
{$ELSE}
            if 0 = AnsiCompareStr(AnsiLowerCase(RealName),
                                  AnsiLowerCase(NodeName))
{$ENDIF}
                then begin
                Result := TempNode;
                if InNetwork and Recursive and //search for a server
                   not IsNetworkServer(TempNode) then
                        Result := nil
                else break;
                        end;
            if Recursive and not IsNetworkServer(TempNode) then begin
                                TempNode.Expand(False);
                Result := FindChild(TempNode,NodeName,True);
                if Result <> nil then break;
                                TempNode.Collapse(True);
            end;
            TempNode := Node.GetNextChild(TempNode);
        end;
    end;
    {**********************************************************}
    function GetFolder(var DirPath: string): string;
    var P: PChar;
    begin
        P := StrScan(PChar(DirPath),'\');
        if P=nil then begin
           Result := Trim(DirPath);
           DirPath := '';
        end
        else begin
             Result := Copy(DirPath,1,P-PChar(Path));
             DirPath := Copy(DirPath,P-PChar(DirPath)+2,StrLen(P));
        end;
    end;
{**********************************************************}
begin

    if FNetBrowse  and  (Path[1]='\') then begin
        Node := GetNetworkRoot;
                if (Path[2]<>'\') or (Node=nil) then Exit;
        Path := Copy(Path,3,999);
        DirName := GetFolder(Path);
        Items.BeginUpdate;
        NodeToExpand := FindChild(Node,DirName,True);
        Items.EndUpdate;
        if NodeToExpand = nil then Exit;
        DirName := GetFolder(Path);
        NodeToExpand.Expand(False);
        NodeToExpand := FindChild(NodeToExpand,DirName,False);
        if NodeToExpand = nil then Exit;
    end
        else NodeToExpand := Items[0];

    TargetNode := NodeToExpand;
    FIsSettingPath := True;

    SetCapture(Handle);
    Screen.Cursor := crHourGlass;

    Items.BeginUpdate;
    NodeToExpand.Collapse(True);
    DirName := GetFolder(Path);
    while True do begin
          if Trim(DirName) = '' then break;
{$IFDEF VER90}
          if NodeToExpand.HasChildren then
                        CanExpand(NodeToExpand);
{$ENDIF}
          NodeToExpand.Expand(False);

          Node := FindChild(NodeToExpand,DirName,False);
          if Node=nil then break;

          TargetNode := Node;
          NodeToExpand := Node;
          DirName := GetFolder(Path);
     end;

    FIsSettingPath := False;
    TargetNode.MakeVisible;
    Selected := TargetNode;
    FDirectory := GetNodePath(Selected);

    NewOnChange(Self,Selected);
    if ExpandNode then Selected.Expand(False);

    Items.EndUpdate;
    Screen.Cursor := crDefault;
    ReleaseCapture;

    FormParent := Parent;
    while FormParent.Parent<>nil do
        FormParent := FormParent.Parent;
    if FormParent.Visible then SetFocus; 
end;

{**********************************************************
                NEW ON COLLAPSED
**********************************************************}
procedure TDirectoryTreeView.NewOnCollapsed(Sender: TObject; Node: TTreeNode);
var Child, NodeToDelete: TTreeNode;
        InNetwork: Boolean;
    DiskFileNotify: TFileNotify;
begin

        if Node.Level > 0 then begin
                InNetwork := IsNetworkNode(Node);
                Child := Node.GetLastChild;
        while Child <> nil do begin
                NodeToDelete := Child;
                Child := Node.GetPrevChild(Child);
            if InNetwork then
                 DeleteNetworkNode(NodeTodelete)
            else Node.Owner.Delete(NodeToDelete);
        end;
        AddDummyChild(Node);
    end
        else begin
                Child := Node.GetLastChild;
        while Child <> nil do begin
                        Child.Collapse(True);
                Child := Node.GetPrevChild(Child);
        end;
    end;

    with Node do begin
         if ImageIndex = OpenFolderIconIndex then
                        ImageIndex := FolderIconIndex;
         if not IsNetworkNode(Node)  and  (Level = 1) then begin
            DiskFileNotify := TFileNotify(Data);
            if (DiskFileNotify <> nil) and FNotification then
               DiskFileNotify.StopMonitoring;
         end;
    end;

    if Assigned(FUserOnCollapsed) then FUserOnCollapsed(Self, Node);
end;

{**********************************************************
                CAN CHANGE
**********************************************************}
function TDirectoryTreeView.CanChange(Node: TTreeNode): Boolean;
begin
        Result := (FFileListBox = nil) or
                  (not FFileListBox.IsRefreshing);
end;

{**********************************************************
                 NEW ON CHANGE
**********************************************************}
procedure TDirectoryTreeView.NewOnChange(
                                        Sender: TObject; Node: TTreeNode);
begin
    FDirectory := GetNodePath(Node);
        SetDirLabelCaption;
    if (FFileListBox <> nil) and (not FIsSettingPath) then
        FFileListBox.Directory := FDirectory;
    if Assigned(FUserOnChange) then FUserOnChange(Self,Node);
end;

{**********************************************************
                CAN EDIT
**********************************************************}
function TDirectoryTreeView.CanEdit(Node: TTreeNode): Boolean;
begin
    Result := FIsEditingNew ;
end;

{**********************************************************
                NOTIFICATION
**********************************************************}
procedure TDirectoryTreeView.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDirLabel then FDirLabel := nil;
    if AComponent = FFileListBox then FFileListBox  := nil;
  end;
end;

{**********************************************************
                ALLOW CREATION
**********************************************************}
function  TDirectoryTreeView.AllowCreation: Boolean;
begin
        Result := False;
    if Selected <> nil then
                if IsNetworkNode(Selected) then
             Result := IsNetworkResource(Selected)
        else Result := (Selected.Level > 0);
end;

{**********************************************************
                FILE NOTIFICATION BLOCK
**********************************************************}
{
  Fnugry File Notify Component (slightly changed)
  Copyright (C) 1997 Gleb Yourchenko
  Version 1.0.1.1

  Contact:
  Gleb Yourchenko
  E-mail: eip__@hotmail.com
}
{ TFileNotify }

procedure TFileNotify.Prepare;
begin
        FThread := TNotificationThread.Create(Self);
end;

procedure TFileNotify.StartMonitoring;
begin
        with FThread do
        if Suspended then Resume;
end;

procedure TFileNotify.StopMonitoring;
begin
        with FThread do
        if not Suspended then Suspend;
end;

constructor TFileNotify.Create(
                AOwner :TCOmponent; AOptions :TNotificationOptions );
begin
  inherited Create;
  FOptions := AOptions; // [fnoFile, fnoFolder];
end;

destructor TFileNotify.Destroy;
begin
        if FThread <> Nil then
      with FThread do begin
        if Suspended then Resume;
        Free;
      end;
        inherited Destroy;
end;

procedure TFileNotify.Notify(Action :TNotificationOption);
begin
  if assigned(FOnNotify) then FOnNotify(Self, Action);
end;

procedure TFileNotify.SetOptions(const Value :TNotificationOptions);
begin
  if Value <> Options then begin
      FOptions := Value;
      if FThread <> Nil then FThread.Reset;
  end;
end;

procedure TFileNotify.SetRootFolder(const Value :String);
begin
  if Value <> FFolder then begin
      FFolder := Value;
      if FFolder[Length(FFolder)] = '\' then
          SetLength(FFolder, Length(FFolder)-1 );
      if FThread <> Nil then FThread.Reset;
  end;
end;

procedure TFileNotify.SetAsyncNotify(Value :Boolean);
begin
  FAsyncNotify := Value;
end;

{ TNotificationThread }

const

  HANDLE_FILE         = 0;
  HANDLE_DIR          = 1;
  HANDLE_ATTR         = 2;
  HANDLE_SIZE         = 3;
  HANDLE_TIME         = 4;

procedure TNotificationThread.Execute;
var R, I :Integer;
begin
  while not Terminated do
    if FHandleCount > 0 then
      begin
        R := WaitForMultipleObjects(FHandleCount, @FHandles, false, 200);
        for I := 0 to FHandleCount-1 do
         if R = (WAIT_OBJECT_0 + I) then
           begin
             FActiveFilter := FHandleFilters[I];
             if FOwner.FAsyncNotify
               then Notify
               else Synchronize(Notify);
             FindNextChangeNotification(FHandles[I]);
             break;
           end;
      end
    else sleep(0);
end;


procedure TNotificationThread.Notify;
begin
  FOwner.Notify(FActiveFilter);
end;

constructor TNotificationThread.Create(Owner :TFileNotify);
begin
  inherited Create(true);
  FOwner := Owner;
  FreeOnTerminate := False;
  Priority := tpLowest;
  Reset;
  Resume;
end;

procedure TNotificationThread.ReleaseHandle;
var I :Integer;
begin
  for I := 0 to FHandleCount-1 do
    if FHandles[I] <> 0 then FindCloseChangeNotification(FHandles[I]);
  fillchar(FHandles, SizeOf(FHandles), 0);
  fillchar(FHandleFilters, Sizeof(FHandleFilters), 0);
  FHandleCount := 0;
end;

destructor TNotificationThread.Destroy;
begin
  ReleaseHandle;
  inherited destroy;
end;

procedure TNotificationThread.AllocateHandle;

  function DoAllocate(Filter :Integer):THandle;
  begin
    if fnoWatchSubTree in FOwner.FOptions then
       result := FindFirstChangeNotification(
         PChar(FOwner.Folder), Bool(1), Filter)
    else
       result := FindFirstChangeNotification(
         PChar(FOwner.Folder), Bool(0), Filter);
    if result = INVALID_HANDLE_VALUE then
      raise EFileNotificationError.Create(SErrAllocHandle);
  end;

begin
  if (FOwner <> Nil) and (FOwner.FOptions <> []) then
    try
      fillchar(FHandles, SizeOf(FHandles), 0);
      FHandleCount := 0;
      if fnoFile in FOwner.FOptions then
        begin
          FHandles[FHandleCount] := DoAllocate(FILE_NOTIFY_CHANGE_FILE_NAME);
          FHandleFilters[FHandleCount] := fnoFile;
          inc(FHandleCount);
        end;
      if fnoFolder in FOwner.FOptions then
        begin
          FHandles[FHandleCount] := DoAllocate(FILE_NOTIFY_CHANGE_DIR_NAME);
          FHandleFilters[FHandleCount] := fnoFolder;
          inc(FHandleCount);
        end;
      if fnoSize in FOwner.FOptions then
        begin
          FHandles[FHandleCount] := DoAllocate(FILE_NOTIFY_CHANGE_SIZE);
          FHandleFilters[FHandleCount] := fnoSize;
          inc(FHandleCount);
        end;
      if fnoTime in FOwner.FOptions then
        begin
          FHandles[FHandleCount] := DoAllocate(FILE_NOTIFY_CHANGE_LAST_WRITE);
          FHandleFilters[FHandleCount] := fnoTime;
          inc(FHandleCount);
        end;
      if fnoAttr in FOwner.FOptions then
        begin
          FHandles[FHandleCount] := DoAllocate(FILE_NOTIFY_CHANGE_ATTRIBUTES);
          FHandleFilters[FHandleCount] := fnoAttr;
          inc(FHandleCount);
        end;
    except
      ReleaseHandle;
//      raise;
    end;
end;

procedure TNotificationThread.Reset;
begin
  ReleaseHandle;
  AllocateHandle;
end;

{**********************************************************
                REGISTER
**********************************************************}
procedure Register;
begin
  RegisterComponents('Win32',
                        [TDirectoryTreeView,TNewFileListBox]);
end;

{**********************************************************
                INIT IMAGE LISTS
**********************************************************}
procedure InitImageLists;
var     Fileinfo: TSHFileInfo;
        PIDL     : PItemIDList;
    F:             TSearchRec;  
begin
    FolderIconIndex := - 1;
    OpenFolderIconIndex := - 1;
    ServerIconIndex := - 1;
    ServerNoAccessIconIndex := - 1;
    NethoodIconIndex := - 1;
    NetworkIconIndex := - 1;
    DomainIconIndex := - 1;

                //     

    SysSmallImageList := TImageList.Create(nil);
        with SysSmallImageList do begin
        Handle := SHGetFileInfo('',
                                0, Fileinfo, sizeof(TSHFileInfo),
                                SHGFI_SMALLICON or SHGFI_SYSICONINDEX );
        ShareImages := True;
    end;
                //    " "

        SHGetSpecialFolderLocation(Application.Handle, CSIDL_DRIVES, PIDL);
        if SHGetFileInfo(PChar(PIDL),
        0, Fileinfo, sizeof(TSHFileInfo),
        SHGFI_PIDL or SHGFI_SYSICONINDEX ) <> 0 then
        MyComputerIcon := Fileinfo.IIcon;

                //    " "

        SHGetSpecialFolderLocation(Application.Handle, CSIDL_NETWORK, PIDL);
        if SHGetFileInfo(PChar(PIDL),
        0, Fileinfo, sizeof(TSHFileInfo),
        SHGFI_PIDL or SHGFI_SYSICONINDEX ) <> 0 then
        NethoodIconIndex := Fileinfo.IIcon;

                //     

        if FindFirstDirectory('C:\*.*', F) then begin
       FolderIconIndex := kpGetImageIndex('C:\'+F.Name);
       FindClose ( F );
    end;

    TreeViewImageList := TImageList.Create(nil);
end;

{**********************************************************
                INITIALIZATION
**********************************************************}
initialization
    InitImageLists;

{**********************************************************
                FINALIZATION
**********************************************************}
finalization
        SysSmallImageList.Free;
        TreeViewImageList.Free;
end.
